import os
import json
import redis
import logging

import predix.service
import predix.config

class Cache(object):
    """
    A simple Key-Value in memory data store (Redis).

    .. important::

       This service will only work from the Predix Cloud -- Firewall will block
       any traffic not originating from within the Predix environment.  If you
       attempt you'll likely see a ConnectionError *Error 60* with *Operation
       timed out.*

    For more information about Predix Cache see the catalog and official
    documentation.

    https://www.predix.io/services/service.html?id=1215

    """
    def __init__(self, host=None, port=None, password=None,
            *args, **kwargs):

        if not predix.config.is_cf_env():
            raise predix.config.PredixCloudRequiredError()

        self.host = host or self._get_host()
        self.port = port or self._get_port()
        self.password = password or self._get_password()

        self.connection = redis.StrictRedis(host=self.host, port=self.port,
                password=self.password)

    def _get_host(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_cache = services['predix-cache'][0]['credentials']
            return predix_cache['host']
        else:
            return predix.config.get_env_value(self, 'host')

    def _get_password(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_cache = services['predix-cache'][0]['credentials']
            return predix_cache['password']
        else:
            return predix.config.get_env_value(self, 'password')

    def _get_port(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_cache = services['predix-cache'][0]['credentials']
            return predix_cache['port']
        else:
            return predix.config.get_env_value(self, 'port')

    def get(self, key):
        """
        Return the value stored for the given key.
        """
        return self.connection.get(key)

    def set(self, key, value):
        """
        Set the given key to the given value.
        """
        return self.connection.set(key, value)




import os
import uuid
import urllib
import logging

import predix.config
import predix.service


class AccessControl(object):
    """
    Use the Access Control service to provide a more powerful authorization
    framework than basic User Account and Authorization (UAA) service.

    Access Control service provides app-specific policies without adding
    overhead to a UAA server that may become the entry point for several apps
    over time.

    :param uri: The URI endpoint for accessing the ACS service instance.

    :param zone_id: The Predix-Zone-Id assigned to the tenant for the ACS
    service instance.

    """
    def __init__(self, uri=None, zone_id=None, *args, **kwargs):
        super(AccessControl, self).__init__(*args, **kwargs)

        self.zone_id = zone_id or self._get_zone_id()
        self.uri = uri or self._get_uri()

        self.service = predix.service.Service(self.zone_id)

    def _get_zone_id(self):
        """
        Returns the Predix Zone Id for the service that is a required
        header in service calls.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_acs = services['predix-acs'][0]['credentials']
            return predix_acs['zone']['http-header-value']
        else:
            return predix.config.get_env_value(self, 'zone_id')

    def _get_uri(self):
        """
        Returns the URI endpoint for an instance of the ACS
        service from environment inspection.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_acs = services['predix-acs'][0]['credentials']
            return predix_acs['uri']
        else:
            return predix.config.get_env_value(self, 'uri')

    def authenticate_as_client(self, client_id, client_secret):
        """
        Will authenticate for the given client / secret.
        """
        self.service.uaa.authenticate(client_id, client_secret)

    def _get_resource_uri(self, guid=None):
        """
        Returns the full path that uniquely identifies
        the resource endpoint.
        """
        uri = self.uri + '/v1/resource'
        if guid:
            uri += '/' + urllib.quote_plus(guid)
        return uri

    def get_resources(self):
        """
        Return all of the resources in the ACS service.
        """
        uri = self._get_resource_uri()
        return self.service._get(uri)

    def get_resource(self, resource_id):
        """
        Returns a specific resource by resource id.
        """
        # resource_id could be a path such as '/asset/123' so quote
        uri = self._get_resource_uri(guid=resource_id)
        return self.service._get(uri)

    def _post_resource(self, body):
        """
        Create new resources and associated attributes.

        Example:

            acs.post_resource([
                {
                    "resourceIdentifier": "masaya",
                    "parents": [],
                    "attributes": [
                        {
                            "issuer": "default",
                            "name": "country",
                            "value": "Nicaragua"
                            }
                        ],
                }
            ])

        The issuer is effectively a namespace, and in policy evaluations you
        identify an attribute by a specific namespace.  Many examples provide
        a URL but it could be any arbitrary string.

        The body is a list, so many resources can be added at the same time.
        """
        assert isinstance(body, (list)), "POST for requires body to be a list"
        uri = self._get_resource_uri()
        return self.service._post(uri, body)

    def delete_resource(self, resource_id):
        """
        Remove a specific resource by its identifier.
        """
        # resource_id could be a path such as '/asset/123' so quote
        uri = self._get_resource_uri(guid=resource_id)
        return self.service._delete(uri)

    def _put_resource(self, resource_id, body):
        """
        Update a resource for the given resource id.  The body is not
        a list but a dictionary of a single resource.
        """
        assert isinstance(body, (dict)), "PUT requires body to be a dict."
        # resource_id could be a path such as '/asset/123' so quote
        uri = self._get_resource_uri(guid=resource_id)
        return self.service._put(uri, body)

    def add_resource(self, resource_id, attributes, parents=[],
            issuer='default'):
        """
        Will add the given resource with a given identifier and attribute
        dictionary.

            example/

                add_resource('/asset/12', {'id': 12, 'manufacturer': 'GE'})
        """
        # MAINT: consider test to avoid adding duplicate resource id
        assert isinstance(attributes, (dict)), "attributes expected to be dict"

        attrs = []
        for key in attributes.keys():
            attrs.append({
                'issuer': issuer,
                'name': key,
                'value': attributes[key]
                })

        body = {
            "resourceIdentifier": resource_id,
            "parents": parents,
            "attributes": attrs,
        }

        return self._put_resource(resource_id, body)

    def _get_subject_uri(self, guid=None):
        """
        Returns the full path that uniquely identifies
        the subject endpoint.
        """
        uri = self.uri + '/v1/subject'
        if guid:
            uri += '/' + urllib.quote_plus(guid)
        return uri

    def get_subjects(self):
        """
        Return all of the subjects in the ACS service.
        """
        uri = self._get_subject_uri()
        return self.service._get(uri)

    def get_subject(self, subject_id):
        """
        Returns a specific subject by subject id.
        """
        # subject_id could be a path such as '/user/j12y' so quote
        uri = self._get_subject_uri(guid=subject_id)
        return self.service._get(uri)

    def _post_subject(self, body):
        """
        Create new subjects and associated attributes.

        Example:

            acs.post_subject([
                {
                    "subjectIdentifier": "/role/evangelist",
                    "parents": [],
                    "attributes": [
                        {
                            "issuer": "default",
                            "name": "role",
                            "value": "developer evangelist",
                        }
                    ]
                }
            ])

        The issuer is effectively a namespace, and in policy evaluations
        you identify an attribute by a specific namespace.  Many examples
        provide a URL but it could be any arbitrary string.

        The body is a list, so many subjects can be added at the same time.
        """
        assert isinstance(body, (list)), "POST requires body to be a list"

        uri = self._get_subject_uri()
        return self.service._post(uri, body)

    def delete_subject(self, subject_id):
        """
        Remove a specific subject by its identifier.
        """
        # subject_id could be a path such as '/role/analyst' so quote
        uri = self._get_subject_uri(guid=subject_id)
        return self.service._delete(uri)

    def _put_subject(self, subject_id, body):
        """
        Update a subject for the given subject id.  The body is not
        a list but a dictionary of a single resource.
        """
        assert isinstance(body, (dict)), "PUT requires body to be dict."

        # subject_id could be a path such as '/asset/123' so quote
        uri = self._get_subject_uri(guid=subject_id)
        return self.service._put(uri, body)

    def add_subject(self, subject_id, attributes, parents=[],
            issuer='default'):
        """
        Will add the given subject with a given identifier and attribute
        dictionary.

            example/

                add_subject('/user/j12y', {'username': 'j12y'})
        """
        # MAINT: consider test to avoid adding duplicate subject id
        assert isinstance(attributes, (dict)), "attributes expected to be dict"

        attrs = []
        for key in attributes.keys():
            attrs.append({
                'issuer': issuer,
                'name': key,
                'value': attributes[key]
                })

        body = {
            "subjectIdentifier": subject_id,
            "parents": parents,
            "attributes": attrs,
        }

        return self._put_subject(subject_id, body)

    def _get_monitoring_heartbeat(self):
        """
        Tests whether or not the ACS service being monitored is alive.
        """
        target = self.uri + '/monitoring/heartbeat'
        response = self.session.get(target)
        return response

    def is_alive(self):
        """
        Will test whether the ACS service is up and alive.
        """
        response = self.get_monitoring_heartbeat()
        if response.status_code == 200 and response.content == 'alive':
            return True

        return False

    def _get_policy_set_uri(self, guid=None):
        """
        Returns the full path that uniquely identifies
        the subject endpoint.
        """
        uri = self.uri + '/v1/policy-set'
        if guid:
            uri += '/' + urllib.quote_plus(guid)
        return uri

    def get_policy_sets(self):
        """
        Return all of the policy sets in the ACS service.
        """
        uri = self._get_policy_set_uri()
        return self.service._get(uri)

    def _put_policy_set(self, policy_set_id, body):
        """
        Will create or update a policy set for the given path.
        """
        assert isinstance(body, (dict)), "PUT requires body to be a dict."
        uri = self._get_policy_set_uri(guid=policy_set_id)
        return self.service._put(uri, body)

    def _get_policy_set(self, policy_set_id):
        """
        Get a specific policy set by id.
        """
        uri = self._get_policy_set_uri(guid=policy_set_id)
        return self.service._get(uri)

    def delete_policy_set(self, policy_set_id):
        """
        Delete a specific policy set by id.  Method is idempotent.
        """
        uri = self._get_policy_set_uri(guid=policy_set_id)
        return self.service._delete(uri)

    def add_policy(self, name, action, resource, subject, condition,
            policy_set_id=None, effect='PERMIT'):
        """
        Will create a new policy set to enforce the given policy details.

        The name is just a helpful descriptor for the policy.

        The action maps to a HTTP verb.

        Policies are evaluated against resources and subjects.  They are
        identified by matching a uriTemplate or attributes.

        Examples::

            resource = {
                "uriTemplate": "/asset/{id}"
                }

            subject: {
                "attributes": [{
                    "issuer": "default",
                    "name": "role"
                    }]
                }

        The condition is expected to be a string that defines a groovy
        operation that can be evaluated.

        Examples::

            condition = "match.single(subject.attributes('default', 'role'),
                'admin')

        """
        # If not given a policy set id will generate one
        if not policy_set_id:
            policy_set_id = str(uuid.uuid4())

        # Only a few operations / actions are supported in policy definitions
        if action not in ['GET', 'PUT', 'POST', 'DELETE']:
            raise ValueError("Invalid action")

        # Defines a single policy to be part of the policy set.
        policy = {
            "name": name,
            "target": {
                "resource": resource,
                "subject": subject,
                "action": action,
                },
            "conditions": [{
                "name": "",
                "condition": condition,
                }],
            "effect": effect,
        }

        # Body of the request is a list of policies
        body = {
            "name": policy_set_id,
            "policies": [policy],
        }

        result = self._put_policy_set(policy_set_id, body)
        return result

    def is_allowed(self, subject_id, action, resource_id, policy_sets=[]):
        """
        Evaluate a policy-set against a subject and resource.

        example/

            is_allowed('/user/j12y', 'GET', '/asset/12')

        """
        body = {
            "action": action,
            "subjectIdentifier": subject_id,
            "resourceIdentifier": resource_id,
        }

        if policy_sets:
            body['policySetsEvaluationOrder'] = policy_sets

        # Will return a 200 with decision
        uri = self.uri + '/v1/policy-evaluation'

        logging.debug("URI=" + str(uri))
        logging.debug("BODY=" + str(body))

        response = self.service._post(uri, body)

        if 'effect' in response:
            if response['effect'] in ['NOT_APPLICABLE', 'PERMIT']:
                return True

        return False




import os
import json
import logging


class CloudFoundryLoginError(Exception):
    """
    A Cloud Foundry Login Error will be raised when the operation
    is not permitted due to an invalid token.
    """
    def __init__(self, message):
        super(CloudFoundryLoginError, self).__init__(message)
        logging.warning("Try 'cf login' to get a new token from Cloud Foundry (%s)." % (message))


class Config(object):
    """
    Interface into a cached cloud foundry configuration file.
    """
    def __init__(self, config_file='~/.cf/config.json', *args, **kwargs):
        super(Config, self).__init__(*args, **kwargs)

        logging.debug("Using CF config file: %s" % (config_file))
        self.config_file = config_file
        self.config = self._get_cloud_foundry_config()

    def _get_cloud_foundry_config(self):
        """
        Reads the local cf CLI cache stored in the users
        home directory.
        """
        config = os.path.expanduser(self.config_file)
        if not os.path.exists(config):
            raise CloudFoundryLoginError('You must run `cf login` to authenticate')

        with open(config, "r") as data:
            return json.load(data)

    def get_access_token(self):
        """
        Returns the access token from the cache, including the
        "Bearer" identifier.
        """
        return self.config['AccessToken']

    def get_uaa_endpoint(self):
        """
        Returns the UAA endpoint used for Predix Cloud Foundry
        authentication.
        """
        return self.config['UaaEndpoint']

    def get_target(self):
        """
        Returns the API target URI.
        """
        return self.config['Target']

    def _get_organization_info(self):
        """
        Returns all cached information about the org.
        """
        return self.config['OrganizationFields']

    def get_organization_name(self):
        """
        Returns the name of the organization currently targeted.
        """
        if 'PREDIX_ORGANIZATION_NAME' in os.environ:
            return os.environ['PREDIX_ORGANIZATION_NAME']
        else:
            return self._get_organization_info()['Name']

    def get_organization_guid(self):
        """
        Returns the GUID for the organization currently targeted.
        """
        if 'PREDIX_ORGANIZATION_GUID' in os.environ:
            return os.environ['PREDIX_ORGANIZATION_GUID']
        else:
            info = self._get_organization_info()
            for key in ('Guid', 'GUID'):
                if key in info.keys():
                    return info[key]
            raise ValueError('Unable to determine cf organization guid')

    def _get_space_info(self):
        """
        Returns all cached information about the space.
        """
        return self.config['SpaceFields']

    def get_space_name(self):
        """
        Returns the name of the space currently targeted.
        """
        if 'PREDIX_SPACE_NAME' in os.environ:
            return os.environ['PREDIX_SPACE_NAME']
        else:
            return self._get_space_info()['Name']

    def get_space_guid(self):
        """
        Returns the GUID for the space currently targeted.

        Can be set by environment variable with PREDIX_SPACE_GUID.
        Can be determined by ~/.cf/config.json.
        """
        if 'PREDIX_SPACE_GUID' in os.environ:
            return os.environ['PREDIX_SPACE_GUID']
        else:
            info = self._get_space_info()
            for key in ('Guid', 'GUID'):
                if key in info.keys():
                    return info[key]
            raise ValueError('Unable to determine cf space guid')




import os

import predix.config
import predix.admin.service
import predix.data.blobstore


class BlobStore(object):
    """
    Blob Store for binary large object storage.
    """
    def __init__(self, plan_name=None, name=None, uaa=None, *args, **kwargs):
        super(BlobStore, self).__init__(*args, **kwargs)
        self.service_name = 'predix-blobstore'
        self.plan_name = plan_name or 'Tiered'
        self.use_class = predix.data.blobstore.BlobStore

        self.service = predix.admin.service.CloudFoundryService(self.service_name,
                self.plan_name, name=name)

    def exists(self):
        """
        Returns whether or not this service already exists.
        """
        return self.service.exists()

    def create(self, **kwargs):
        """
        Create an instance of the Blob Store Service with the typical
        starting settings.
        """
        self.service.create(**kwargs)

        predix.config.set_env_value(self.use_class, 'url',
                self.service.settings.data['url'])
        predix.config.set_env_value(self.use_class, 'access_key_id',
                self.service.settings.data['access_key_id'])
        predix.config.set_env_value(self.use_class, 'bucket_name',
                self.service.settings.data['bucket_name'])
        predix.config.set_env_value(self.use_class, 'host',
                self.service.settings.data['host'])
        predix.config.set_env_value(self.use_class, 'secret_access_key',
                self.service.settings.data['secret_access_key'])

    def add_to_manifest(self, manifest):
        """
        Add useful details to the manifest about this service
        so that it can be used in an application.

        :param manifest: An predix.admin.app.Manifest object
            instance that manages reading/writing manifest config
            for a cloud foundry app.
        """
        # Add this service to the list of services
        manifest.add_service(self.service.name)

        # Add environment variables

        url = predix.config.get_env_key(self.use_class, 'url')
        manifest.add_env_var(url, self.service.settings.data['url'])

        akid = predix.config.get_env_key(self.use_class, 'access_key_id')
        manifest.add_env_var(akid, self.service.settings.data['access_key_id'])

        bucket = predix.config.get_env_key(self.use_class, 'bucket_name')
        manifest.add_env_var(bucket, self.service.settings.data['bucket_name'])

        host = predix.config.get_env_key(self.use_class, 'host')
        manifest.add_env_var(host, self.service.settings.data['host'])

        secret_access_key = predix.config.get_env_key(self.use_class, 'secret_access_key')
        manifest.add_env_var(secret_access_key, self.service.settings.data['secret_access_key'])

        manifest.write_manifest()

    def enable_encryption(self):
        os.environ['ENABLE_SERVER_SIDE_ENCRYPTION'] = True




import os
import yaml
import copy
import logging
from cryptography.fernet import Fernet
from builtins import bytes

import predix.config


class Manifest(object):
    """
    Cloud Foundry utilizes a MANIFEST.yml file as the source
    of application configuration.  As we setup services and
    run our applications the manifest is a place to store
    important configuration details.

    :param app_name: the name of your application which will be used by default
        in the route
    :param manifest_key: if encrypting your manifest this is the key for
        cryptography
    :param encrypted: whether to manifest values should be encrypted
    :param debug: enable additional debugging

    """
    def __init__(self, manifest_path='manifest.yml',
            app_name='my-predix-app',
            manifest_key='~/.predix/manifest_key',
            encrypted=False,
            debug=False):

        self.manifest_path = os.path.expanduser(manifest_path)
        self.app_name = app_name

        # Parameters for encrypting config files
        self.manifest_key = manifest_key
        self.encrypted = encrypted

        if not self.encrypted:
            logging.warning("Writing manifest {} unencrypted.".format(manifest_path))

        # App may have a client
        self._client_id = None
        self._client_secret = None

        if debug:
            logging.basicConfig(level=logging.DEBUG)

        # Read or Generate a manifest file
        if os.path.exists(self.manifest_path):
            self.read_manifest()
        else:
            self.create_manifest()

        # Probably always want manifest loaded into environment
        self.set_os_environ()

    def get_manifest_version(self):
        """
        Returns the version of PredixPy used to generate the manifest.
        """
        if 'env' in self.manifest:
            if 'PREDIXPY_VERSION' in self.manifest['env']:
                return self.manifest['env']['PREDIXPY_VERSION']

        return None

    def read_manifest(self, encrypted=None):
        """
        Read an existing manifest.
        """
        with open(self.manifest_path, 'r') as input_file:
            self.manifest = yaml.safe_load(input_file)
            if 'env' not in self.manifest:
                self.manifest['env'] = {}
            if 'services' not in self.manifest:
                self.manifest['services'] = []

            # If manifest is encrypted, use manifest key to
            # decrypt each value before storing in memory.

            if 'PREDIXPY_ENCRYPTED' in self.manifest['env']:
                self.encrypted = True

            if encrypted or self.encrypted:
                key = predix.config.get_crypt_key(self.manifest_key)
                f = Fernet(key)

                for var in self.manifest['env'].keys():
                    value = f.decrypt(bytes(self.manifest['env'][var], 'utf-8'))
                    self.manifest['env'][var] = value.decode('utf-8')

            self.app_name = self.manifest['applications'][0]['name']

            input_file.close()

    def create_manifest(self):
        """
        Create a new manifest and write it to
        disk.
        """
        self.manifest = {}
        self.manifest['applications'] = [{'name': self.app_name}]
        self.manifest['services'] = []
        self.manifest['env'] = {
                'PREDIXPY_VERSION': str(predix.version),
                }

        self.write_manifest()

    def _get_encrypted_manifest(self):
        """
        Returns contents of the manifest where environment variables
        that are secret will be encrypted without modifying the existing
        state in memory which will remain unencrypted.
        """
        key = predix.config.get_crypt_key(self.manifest_key)
        f = Fernet(key)

        manifest = copy.deepcopy(self.manifest)
        for var in self.manifest['env'].keys():
            value = str(self.manifest['env'][var])
            manifest['env'][var] = f.encrypt(bytes(value, 'utf-8')).decode('utf-8')

        return manifest

    def write_manifest(self, manifest_path=None, encrypted=None):
        """
        Write manifest to disk.

        :param manifest_path: write to a different location
        :param encrypted: write with env data encrypted

        """
        manifest_path = manifest_path or self.manifest_path
        self.manifest['env']['PREDIXPY_VERSION'] = str(predix.version)

        with open(manifest_path, 'w') as output_file:
            if encrypted or self.encrypted:
                self.manifest['env']['PREDIXPY_ENCRYPTED'] = self.manifest_key
                content = self._get_encrypted_manifest()
            else:
                content = self.manifest   # shallow reference
                if 'PREDIXPY_ENCRYPTED' in content['env']:
                    del(content['env']['PREDIXPY_ENCRYPTED'])

            yaml.safe_dump(content, output_file,
                    default_flow_style=False, explicit_start=True)
            output_file.close()

    def add_env_var(self, key, value):
        """
        Add the given key / value as another environment
        variable.
        """
        self.manifest['env'][key] = value
        os.environ[key] = str(value)

    def add_service(self, service_name):
        """
        Add the given service to the manifest.
        """
        if service_name not in self.manifest['services']:
            self.manifest['services'].append(service_name)

    def set_os_environ(self):
        """
        Will load any environment variables found in the
        manifest file into the current process for use
        by applications.

        When apps run in cloud foundry this would happen
        automatically.
        """
        for key in self.manifest['env'].keys():
            os.environ[key] = str(self.manifest['env'][key])

    def get_client_id(self):
        """
        Return the client id that should have all the
        needed scopes and authorities for the services
        in this manifest.
        """
        self._client_id = predix.config.get_env_value(predix.app.Manifest, 'client_id')
        return self._client_id

    def get_client_secret(self):
        """
        Return the client secret that should correspond with
        the client id.
        """
        self._client_secret = predix.config.get_env_value(predix.app.Manifest, 'client_secret')
        return self._client_secret

    def get_timeseries(self, *args, **kwargs):
        """
        Returns an instance of the Time Series Service.
        """
        import predix.data.timeseries
        ts = predix.data.timeseries.TimeSeries(*args, **kwargs)
        return ts

    def get_asset(self):
        """
        Returns an instance of the Asset Service.
        """
        import predix.data.asset
        asset = predix.data.asset.Asset()
        return asset

    def get_uaa(self):
        """
        Returns an insstance of the UAA Service.
        """
        import predix.security.uaa
        uaa = predix.security.uaa.UserAccountAuthentication()
        return uaa

    def get_acs(self):
        """
        Returns an instance of the Asset Control Service.
        """
        import predix.security.acs
        acs = predix.security.acs.AccessControl()
        return acs

    def get_weather(self):
        """
        Returns an instance of the Weather Service.
        """
        import predix.data.weather
        weather = predix.data.weather.WeatherForecast()
        return weather

    def get_blobstore(self):
        import predix.data.blobstore
        blobstore = predix.data.blobstore.BlobStore()
        return blobstore

    def get_cache(self):
        import predix.data.cache
        cache = predix.data.cache.Cache()
        return cache

    def get_dbaas(self, **kwargs):
        import predix.data.dbaas
        pg = predix.data.dbaas.PostgreSQL(**kwargs)
        return pg

    def get_eventhub(self, publish_config=None, subscribe_config=None):
        import predix.data.eventhub.client
        eventhub = predix.data.eventhub.client.Eventhub(subscribe_config=subscribe_config,
                                                        publish_config=publish_config)
        return eventhub




import logging

import predix.app
import predix.admin.uaa
import predix.admin.acs
import predix.admin.asset
import predix.admin.weather
import predix.admin.cf.spaces
import predix.admin.blobstore
import predix.admin.timeseries
import predix.admin.logstash
import predix.admin.cache
import predix.admin.dbaas


class Manifest(predix.app.Manifest):
    """
    Extends Application Manifest with administrative
    functions for creating and configuring services.
    """
    def __init__(self, *args, **kwargs):
        super(Manifest, self).__init__(*args, **kwargs)

        self.space = predix.admin.cf.spaces.Space()
        self.services = predix.admin.cf.services.Service()

        self.supported = {
            'predix-uaa': predix.admin.uaa.UserAccountAuthentication,
            'predix-acs': predix.admin.acs.AccessControl,
            'predix-asset': predix.admin.asset.Asset,
            'predix-blobstore': predix.admin.blobstore.BlobStore,
            'predix-cache': predix.admin.cache.Cache,
            'predix-dbaas': predix.admin.dbaas.PostgreSQL,
            'predix-timeseries': predix.admin.timeseries.TimeSeries,
            'predix-weather': predix.admin.weather.WeatherForecast,
            'predix-logging': predix.admin.logstash.Logging,
        }

    def create_manifest_from_space(self):
        """
        Populate a manifest file generated from details from the
        cloud foundry space environment.
        """
        space = predix.admin.cf.spaces.Space()

        summary = space.get_space_summary()
        for instance in summary['services']:
            service_type = instance['service_plan']['service']['label']
            name = instance['name']
            if service_type in self.supported:
                service = self.supported[service_type](name=name)
                service.add_to_manifest(self)
            elif service_type == 'us-weather-forecast':
                weather = predix.admin.weather.WeatherForecast(name=name)
                weather.add_to_manifest(self)
            else:
                logging.warning("Unsupported service type: %s" % service_type)

    def lock_to_org_space(self):
        """
        Lock the manifest to the current organization and space regardless of
        Cloud Foundry target.
        """
        self.add_env_var('PREDIX_ORGANIZATION_GUID', self.space.org.guid)
        self.add_env_var('PREDIX_ORGANIZATION_NAME', self.space.org.name)
        self.add_env_var('PREDIX_SPACE_GUID', self.space.guid)
        self.add_env_var('PREDIX_SPACE_NAME', self.space.name)
        self.write_manifest()

    def create_uaa(self, admin_secret, **kwargs):
        """
        Creates an instance of UAA Service.

        :param admin_secret: The secret password for administering the service
            such as adding clients and users.
        """
        uaa = predix.admin.uaa.UserAccountAuthentication(**kwargs)
        if not uaa.exists():
            uaa.create(admin_secret, **kwargs)

        uaa.add_to_manifest(self)
        return uaa

    def create_client(self, client_id=None, client_secret=None, uaa=None):
        """
        Create a client and add it to the manifest.

        :param client_id: The client id used to authenticate as a client
            in UAA.

        :param client_secret: The secret password used by a client to
            authenticate and generate a UAA token.

        :param uaa: The UAA to create client with
        """
        if not uaa:
            uaa = predix.admin.uaa.UserAccountAuthentication()

        # Client id and secret can be generated if not provided as arguments

        if not client_id:
            client_id = uaa._create_id()

        if not client_secret:
            client_secret = uaa._create_secret()

        uaa.create_client(client_id, client_secret)
        uaa.add_client_to_manifest(client_id, client_secret, self)

    def create_timeseries(self, **kwargs):
        """
        Creates an instance of the Time Series Service.
        """
        ts = predix.admin.timeseries.TimeSeries(**kwargs)
        ts.create()

        client_id = self.get_client_id()
        if client_id:
            ts.grant_client(client_id)

        ts.add_to_manifest(self)
        return ts

    def create_asset(self, **kwargs):
        """
        Creates an instance of the Asset Service.
        """
        asset = predix.admin.asset.Asset(**kwargs)
        asset.create()

        client_id = self.get_client_id()
        if client_id:
            asset.grant_client(client_id)

        asset.add_to_manifest(self)
        return asset

    def create_acs(self, **kwargs):
        """
        Creates an instance of the Asset Service.
        """
        acs = predix.admin.acs.AccessControl(**kwargs)
        acs.create()

        client_id = self.get_client_id()
        if client_id:
            acs.grant_client(client_id)

        acs.grant_client(client_id)
        acs.add_to_manifest(self)
        return acs

    def create_weather(self, **kwargs):
        """
        Creates an instance of the Asset Service.
        """
        weather = predix.admin.weather.WeatherForecast(**kwargs)
        weather.create()

        client_id = self.get_client_id()
        if client_id:
            weather.grant_client(client_id)

        weather.grant_client(client_id)
        weather.add_to_manifest(self)
        return weather

    def create_blobstore(self, **kwargs):
        """
        Creates an instance of the BlobStore Service.
        """
        blobstore = predix.admin.blobstore.BlobStore(**kwargs)
        blobstore.create()

        blobstore.add_to_manifest(self)
        return blobstore

    def create_logstash(self, **kwargs):
        """
        Creates an instance of the Logging Service.
        """
        logstash = predix.admin.logstash.Logging(**kwargs)
        logstash.create()
        logstash.add_to_manifest(self)

        logging.info('Install Kibana-Me-Logs application by following GitHub instructions')
        logging.info('git clone https://github.com/cloudfoundry-community/kibana-me-logs.git')

        return logstash

    def create_cache(self, **kwargs):
        """
        Creates an instance of the Cache Service.
        """
        cache = predix.admin.cache.Cache(**kwargs)
        cache.create(**kwargs)
        cache.add_to_manifest(self)
        return cache

    def create_dbaas(self, **kwargs):
        """
        """
        pg = predix.admin.dbaas.PostgreSQL(**kwargs)
        pg.create()
        pg.add_to_manifest(self)
        return pg

    def create_eventhub(self, **kwargs):
        """
        todo make it so the client can be customised to publish/subscribe
        Creates an instance of eventhub service
        """
        eventhub = predix.admin.eventhub.EventHub(**kwargs)
        eventhub.create()
        eventhub.add_to_manifest(self)
        eventhub.grant_client(client_id=self.get_client_id(), **kwargs)
        eventhub.add_to_manifest(self)
        return eventhub


    def get_service_marketplace(self, available=True, unavailable=False,
            deprecated=False):
        """
        Returns a list of service names.  Can return all services, just
        those supported by PredixPy, or just those not yet supported by
        PredixPy.

        :param available: Return the services that are
            available in PredixPy.  (Defaults to True)

        :param unavailable: Return the services that are not yet
            supported by PredixPy.  (Defaults to False)

        :param deprecated: Return the services that are
            supported by PredixPy but no longer available. (True)
        """

        supported = set(self.supported.keys())
        all_services = set(self.space.get_services())

        results = set()
        if available:
            results.update(supported)
        if unavailable:
            results.update(all_services.difference(supported))
        if deprecated:
            results.update(supported.difference(all_services))

        return list(results)




import os

import predix.config
import predix.admin.service
import predix.admin.cf.spaces

class Logging(object):
    """
    An Elasticsearch ELK distribution to centralize Cloud Foundry log
    aggregation and analysis.

    You need to install the Kibana-Me-Logs application separately to
    make use of the logging service.
    https://docs.predix.io/en-US/content/service/operations/logging/
    """
    def __init__(self, plan_name=None, name=None, *args, **kwargs):
        super(Logging, self).__init__(*args, **kwargs)

        # Service name differs depending on org/space configuration
        self.service_name = self._find_service_name()
        self.plan_name = plan_name or 'free'

        self.service = predix.admin.service.CloudFoundryService(self.service_name, 
                self.plan_name, name=name)

    def _find_service_name(self):
        """
        For cloud operations there is support for multiple pools of resources
        dedicated to logstash.  The service name as a result follows the
        pattern logstash-{n} where n is some number.  We can find it from the
        service marketplace.
        """
        space = predix.admin.cf.spaces.Space()
        services = space.get_services()
        for service in services:
            if service.startswith('logstash'):
                return service

        return 'logstash-3'

    def exists(self):
        """
        Returns whether or not this service already exists.
        """
        return self.service.exists()

    def create(self):
        """
        Creates an instance of the logstash pipeline.
        """
        self.service.create(create_keys=False)

    def add_to_manifest(self, manifest):
        """
        Add to the manifest to make sure it is bound to the
        application.
        """
        manifest.add_service(self.service.name)
        manifest.write_manifest()




import os
import json
import logging
import sqlalchemy

import predix.service
import predix.config

class PostgreSQL(object):
    """
    A Relational Database Management System on demand (PostgreSQL).

    .. important::

       This service will only work from the Predix Cloud -- Firewall will block
       any traffic not originating from within the Predix environment.  If you
       attempt you'll likely see a ConnectionError *Error 60* with *Operation
       timed out.*

    For more information about Predix Database as a Service see the catalog and
    official documentation.

    https://www.predix.io/services/service.html?id=1178
    """
    def __init__(self, hostname=None, port=None, username=None, password=None,
            database=None, *args, **kwargs):

        if not predix.config.is_cf_env():
            raise predix.config.PredixCloudRequiredError()

        self.hostname = hostname or self._get_hostname()
        self.port = port or self._get_port()
        self.username = username or self._get_username()
        self.password = password or self._get_password()
        self.database = database or 'postgres'

        connection_str = 'postgresql://{}:{}@{}:{}/{}'.format(
                self.username,
                self.password,
                self.hostname,
                self.port,
                self.database,
                )
        self.engine = sqlalchemy.create_engine(connection_str)

    def __del__(self):
        if self.engine:
            # self.engine.close()
            pass

    def _get_hostname(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            postgres = services['postgres-2.0'][0]['credentials']
            return postgres['hostname']
        else:
            return predix.config.get_env_value(self, 'hostname')

    def _get_port(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            postgres = services['postgres-2.0'][0]['credentials']
            return postgres['port']
        else:
            return predix.config.get_env_value(self, 'port')

    def _get_username(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            postgres = services['postgres-2.0'][0]['credentials']
            return postgres['username']
        else:
            return predix.config.get_env_value(self, 'username')

    def _get_password(self):
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            postgres = services['postgres-2.0'][0]['credentials']
            return postgres['password']
        else:
            return predix.config.get_env_value(self, 'password')

    def execute(self, statement, *args, **kwargs):
        """
        This convenience method will execute the query passed in as is.  For
        more complex functionality you may want to use the sqlalchemy engine
        directly, but this serves as an example implementation.

        :param select_query: SQL statement to execute that will identify the
            resultset of interest.

        """
        with self.engine.connect() as conn:
            s = sqlalchemy.sql.text(statement)
            return conn.execute(s, **kwargs)



from predix.data.eventhub import EventHub_pb2, EventHub_pb2_grpc
from predix.data.eventhub.client import Eventhub


class SubscribeConfig:
    class Recency:
        def __init__(self):
            pass

        OLDEST = "Oldest"
        NEWEST = "Newest"

    def __init__(self,
                 subscriber_name='predix_py_subscriber',
                 batching_enabled=False,
                 auto_send_acks=False,
                 batch_size=100,
                 batch_interval_millis=10000,
                 acks_enabled=False,
                 recency=Recency.OLDEST,
                 ack_duration_before_retry_seconds=30,
                 ack_max_retries=10,
                 ack_retry_interval_seconds=30,
                 topics=None):
        """
        Subscribe Config
        :param subscriber_name: The name of the subscriber
        :param batching_enabled: Should the messages be delivered in batches
        :param batch_size: If batching, what should be the size of the batches
        :param batch_interval_millis: If batching, what should be the max interval
        :param acks_enabled: should the service be expecting acks on delivered messages
        :param ack_duration_before_retry_seconds:  How long should the service wait for an ack before it retry
        :param ack_max_retries: How many retries should the service wait for
        :param ack_retry_interval_seconds: after the initial retry, what should be the period of the message retry
        :param recency: What messages should be sent when connected, all messages in the queue or only new messages
        :param topics: What topics should be subscribed too
        """
        self.subscriber_name = subscriber_name
        self.batching_enabled = batching_enabled
        self.batch_size = batch_size
        self.batch_interval_millis = batch_interval_millis
        self.acks_enabled = acks_enabled
        self.recency = recency
        self.ack_duration_before_retry_seconds = ack_duration_before_retry_seconds
        self.ack_max_retries = ack_max_retries
        self.ack_retry_interval_seconds = ack_retry_interval_seconds
        self.topics = topics if topics is not None else []
        if topics is not None:
            raise
class Subscriber:
    def __init__(self, eventhub_client, config, channel):
        self.eventhub_client = eventhub_client
        self._config = config
        self._stub = EventHub_pb2_grpc.SubscriberStub(channel=channel)

        tx_stream = True
        initial_message = None
        if self._config.batching_enabled:
            stub_call = self._stub.subscribe
        elif self._config.acks_enabled:
            stub_call = self._stub.receiveWithAcks
        else:
            stub_call = self._stub.receive
            tx_stream = False
            initial_message = EventHub_pb2.SubscriptionRequest(subscriber=self._config.subscriber_name,
                                                               zone_id=self.eventhub_client.zone_id,
                                                               instance_id='predixpy-subscriber')
        self._rx_messages = []
        self.active = True
        self.run_subscribe_generator = True
        self.grpc_manager = Eventhub.GrpcManager(stub_call=stub_call,
                                                 on_msg_callback=self._subscriber_callback,
                                                 metadata=self._generate_subscribe_headers(),
                                                 tx_stream=tx_stream,
                                                 initial_message=initial_message
                                                 )

    def __del__(self):
        self.grpc_manager.stop_generator()
        self.run_subscribe_generator = False
        self.active = False

    def shutdown(self):
        if self.active:
            self.active = False
            self.grpc_manager.stop_generator()
            self.run_subscribe_generator = False

    def _subscriber_callback(self, rx_message):
        """
        subscriber callback for the GRPC manager, appends them onto the queue
        :param rx_message: SubscriptionMessage or Message
        :return: None
        """
        self._rx_messages.append(rx_message)

    def subscribe(self):
        """
        return a generator for all subscribe messages
        :return: None
        """
        while self.run_subscribe_generator:
            if len(self._rx_messages) != 0:
                yield self._rx_messages.pop(0)
        return

    def send_acks(self, message):
        """
        send acks to the service
        :param message: EventHub_pb2.Message
        :return: None
        """
        if isinstance(message, EventHub_pb2.Message):
            ack = EventHub_pb2.Ack(partition=message.partition, offset=message.offset)
            self.grpc_manager.send_message(EventHub_pb2.SubscriptionResponse(ack=ack))

        elif isinstance(message, EventHub_pb2.SubscriptionMessage):
            acks = []
            for m in message.messages:
                acks.append(EventHub_pb2.Ack(parition=m.partition, offset=m.offset))
            self.grpc_manager.send_message(EventHub_pb2.SubscriptionAcks(ack=acks))

    def _generate_subscribe_headers(self):
        """
        generate the subscribe stub headers based on the supplied config
        :return: i
        """
        headers =[]
        headers.append(('predix-zone-id', self.eventhub_client.zone_id))

        token = self.eventhub_client.service._get_bearer_token()
        headers.append(('subscribername', self._config.subscriber_name))
        headers.append(('authorization', token[(token.index(' ') + 1):]))

        if self._config.topics is []:
            headers.append(('topic', self.eventhub_client.zone_id + '_topic'))
        else:
            for topic in self._config.topics:
                headers.append(('topic', topic))

        headers.append(('offset-newest', str(self._config.recency == self._config.Recency.NEWEST).lower()))

        headers.append(('acks', str(self._config.acks_enabled).lower()))
        if self._config.acks_enabled:
            headers.append(('max-retries', str(self._config.ack_max_retries)))
            headers.append(('retry-interval', str(self._config.ack_retry_interval_seconds) + 's'))
            headers.append(('duration-before-retry', str(self._config.ack_duration_before_retry_seconds) + 's'))

        if self._config.batching_enabled:
            headers.append(('batch-size', str(self._config.batch_size)))
            headers.append(('batch-interval', str(self._config.batch_interval_millis) + 'ms'))

        return headers


def __del__(self):
    """
    Destructor to make sure an open web socket connection is closed.
    """
    self.shutdown()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import EventHub_pb2 as EventHub__pb2


class PublisherStub(object):
  """Publisher service definition
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.send = channel.stream_stream(
        '/predix.eventhub.Publisher/send',
        request_serializer=EventHub__pb2.PublishRequest.SerializeToString,
        response_deserializer=EventHub__pb2.PublishResponse.FromString,
        )


class PublisherServicer(object):
  """Publisher service definition
  """

  def send(self, request_iterator, context):
    """Send message to topic. Topic name will be in URI
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PublisherServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'send': grpc.stream_stream_rpc_method_handler(
          servicer.send,
          request_deserializer=EventHub__pb2.PublishRequest.FromString,
          response_serializer=EventHub__pb2.PublishResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'predix.eventhub.Publisher', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class SubscriberStub(object):
  """Subscriber service definition
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.receive = channel.unary_stream(
        '/predix.eventhub.Subscriber/receive',
        request_serializer=EventHub__pb2.SubscriptionRequest.SerializeToString,
        response_deserializer=EventHub__pb2.Message.FromString,
        )
    self.receiveWithAcks = channel.stream_stream(
        '/predix.eventhub.Subscriber/receiveWithAcks',
        request_serializer=EventHub__pb2.SubscriptionResponse.SerializeToString,
        response_deserializer=EventHub__pb2.Message.FromString,
        )
    self.subscribe = channel.stream_stream(
        '/predix.eventhub.Subscriber/subscribe',
        request_serializer=EventHub__pb2.SubscriptionAcks.SerializeToString,
        response_deserializer=EventHub__pb2.SubscriptionMessage.FromString,
        )


class SubscriberServicer(object):
  """Subscriber service definition
  """

  def receive(self, request, context):
    """Receive message from a topic, as a subscriber
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def receiveWithAcks(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def subscribe(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_SubscriberServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'receive': grpc.unary_stream_rpc_method_handler(
          servicer.receive,
          request_deserializer=EventHub__pb2.SubscriptionRequest.FromString,
          response_serializer=EventHub__pb2.Message.SerializeToString,
      ),
      'receiveWithAcks': grpc.stream_stream_rpc_method_handler(
          servicer.receiveWithAcks,
          request_deserializer=EventHub__pb2.SubscriptionResponse.FromString,
          response_serializer=EventHub__pb2.Message.SerializeToString,
      ),
      'subscribe': grpc.stream_stream_rpc_method_handler(
          servicer.subscribe,
          request_deserializer=EventHub__pb2.SubscriptionAcks.FromString,
          response_serializer=EventHub__pb2.SubscriptionMessage.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'predix.eventhub.Subscriber', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))




import os
import json
import errno
import base64
import logging
import requests
import datetime
import dateutil.parser

import predix.app
import predix.config


class UserAccountAuthentication(object):
    """
    The UAA service manages user account authorization and access control for
    Predix service calls through client credentials captured in a bearer token.

    :param uri: URI for the UAA endpoint to interact with, can be derived from
        environment PREDIX_SECURITY_UAA_URI variable when not specified.

    Useful documentation about interacting with API - https://docs.cloudfoundry.org/api/uaa

    And more documentation - https://github.com/cloudfoundry/uaa/tree/master/docs
    """
    def __init__(self, uri=None, *args, **kwargs):
        super(UserAccountAuthentication, self).__init__(*args, **kwargs)

        self.uri = uri or self._get_uaa_uri()
        self.session = requests.Session()
        self.authenticated = False
        self.client = {}

    def _get_uaa_uri(self):
        """
        Returns the URI endpoint for an instance of a UAA
        service instance from environment inspection.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_uaa = services['predix-uaa'][0]['credentials']
            return predix_uaa['uri']
        else:
            return predix.config.get_env_value(self, 'uri')

    def _authenticate_client(self, client, secret):
        """
        Returns response of authenticating with the given client and
        secret.
        """
        client_s = str.join(':', [client, secret])
        credentials = base64.b64encode(client_s.encode('utf-8')).decode('utf-8')

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Cache-Control': 'no-cache',
            'Authorization': 'Basic ' + credentials
            }
        params = {
            'client_id': client,
            'grant_type': 'client_credentials'
            }
        uri = self.uri + '/oauth/token'

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))
        logging.debug("BODY=" + str(params))

        response = requests.post(uri, headers=headers, params=params)
        if response.status_code == 200:
            logging.debug("RESPONSE=" + str(response.json()))
            return response.json()
        else:
            logging.warning("Failed to authenticate as %s" % (client))
            response.raise_for_status()

    def _authenticate_user(self, user, password):
        """
        Returns the response of authenticating with the given
        user and password.
        """
        headers = self._get_headers()
        params = {
                'username': user,
                'password': password,
                'grant_type': 'password',
                }
        uri = self.uri + '/oauth/token'

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))
        logging.debug("BODY=" + str(params))

        response = requests.post(uri, headers=headers, params=params)
        if response.status_code == 200:
            logging.debug("RESPONSE=" + str(response.json()))
            return response.json()
        else:
            logging.warning("Failed to authenticate %s" % (user))
            response.raise_for_status()

    def is_expired_token(self, client):
        """
        For a given client will test whether or not the token
        has expired.

        This is for testing a client object and does not look up
        from client_id.  You can use _get_client_from_cache() to
        lookup a client from client_id.
        """
        if 'expires' not in client:
            return True

        expires = dateutil.parser.parse(client['expires'])
        if expires < datetime.datetime.now():
            return True

        return False

    def _initialize_uaa_cache(self):
        """
        If we don't yet have a uaa cache we need to
        initialize it.  As there may be more than one
        UAA instance we index by issuer and then store
        any clients, users, etc.
        """
        try:
            os.makedirs(os.path.dirname(self._cache_path))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

        data = {}
        data[self.uri] = []

        return data

    def _read_uaa_cache(self):
        """
        Read cache of UAA client/user details.
        """
        self._cache_path = os.path.expanduser('~/.predix/uaa.json')
        if not os.path.exists(self._cache_path):
            return self._initialize_uaa_cache()

        with open(self._cache_path, 'r') as data:
            return json.load(data)

    def _get_client_from_cache(self, client_id):
        """
        For the given client_id return what is
        cached.
        """
        data = self._read_uaa_cache()

        # Only if we've cached any for this issuer
        if self.uri not in data:
            return

        for client in data[self.uri]:
            if client['id'] == client_id:
                return client

    def _write_to_uaa_cache(self, new_item):
        """
        Cache the client details into a cached file on disk.
        """
        data = self._read_uaa_cache()

        # Initialize client list if first time
        if self.uri not in data:
            data[self.uri] = []

        # Remove existing client record and any expired tokens
        for client in data[self.uri]:
            if new_item['id'] == client['id']:
                data[self.uri].remove(client)
                continue

            # May have old tokens laying around to be cleaned up
            if 'expires' in client:
                expires = dateutil.parser.parse(client['expires'])
                if expires < datetime.datetime.now():
                    data[self.uri].remove(client)
                    continue

        data[self.uri].append(new_item)

        with open(self._cache_path, 'w') as output:
            output.write(json.dumps(data, sort_keys=True, indent=4))

    def authenticate(self, client_id, client_secret, use_cache=True):
        """
        Authenticate the given client against UAA.  The resulting token
        will be cached for reuse.
        """
        # We will reuse a token for as long as we have one cached
        # and it hasn't expired.
        if use_cache:
            client = self._get_client_from_cache(client_id)
            if (client) and (not self.is_expired_token(client)):
                self.authenticated = True
                self.client = client
                return

        # Let's authenticate the client
        client = {
            'id': client_id,
            'secret': client_secret
        }

        res = self._authenticate_client(client_id, client_secret)
        client.update(res)

        expires = datetime.datetime.now() + \
                  datetime.timedelta(seconds=res['expires_in'])
        client['expires'] = expires.isoformat()

        # Cache it for repeated use until expired
        self._write_to_uaa_cache(client)

        self.client = client
        self.authenticated = True

    def logout(self):
        """
        Log currently authenticated user out, invalidating any existing tokens.
        """
        # Remove token from local cache
        # MAINT: need to expire token on server
        data = self._read_uaa_cache()
        if self.uri in data:
            for client in data[self.uri]:
                if client['id'] == self.client['id']:
                    data[self.uri].remove(client)

        with open(self._cache_path, 'w') as output:
            output.write(json.dumps(data, sort_keys=True, indent=4))

    def _get_headers(self):
        """
        Returns headers needed for UAA operations.
        """
        headers = {
            "pragma": "no-cache",
            "Cache-Control": "no-cache",
            "Content-Type": "application/json",
            "Accepts": "application/json",
            "Authorization": "Bearer " + self.get_token()
        }
        return headers

    def _get(self, uri, params=None, headers=None):
        """
        Simple GET request for a given uri path.
        """
        if not headers:
            headers = self._get_headers()

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))

        response = self.session.get(uri, headers=headers, params=params)
        logging.debug("STATUS=" + str(response.status_code))
        if response.status_code == 200:
            return response.json()
        else:
            logging.error(b"ERROR=" + response.content)
            response.raise_for_status()

    def _post(self, uri, data, headers=None):
        """
        Simple POST request for a given uri path.
        """
        if not headers:
            headers = self._get_headers()

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))
        logging.debug("BODY=" + str(data))

        response = self.session.post(uri, headers=headers,
                data=json.dumps(data))

        logging.debug("STATUS=" + str(response.status_code))
        if response.status_code in [200, 201]:
            return response.json()
        else:
            logging.error(b"ERROR=" + response.content)
            response.raise_for_status()

    def get_token(self):
        """
        Returns the bare access token for the authorized client.
        """
        if not self.authenticated:
            raise ValueError("Must authenticate() as a client first.")

        # If token has expired we'll need to refresh and get a new
        # client credential
        if self.is_expired_token(self.client):
            logging.info("client token expired, will need to refresh token")
            self.authenticate(self.client['id'], self.client['secret'],
                    use_cache=False)

        return self.client['access_token']

    def get_scopes(self):
        """
        Returns the scopes for the authenticated client.
        """
        if not self.authenticated:
            raise ValueError("Must authenticate() as a client first.")

        scope = self.client['scope']
        return scope.split()

    def get_authorization_headers(self):
        """
        Returns the authorization headers with the bearer token needed for
        making calls to Predix Services protected by UAA.
        """
        return {
                'Authorization': 'Bearer ' + self.get_token()
                }

    def is_admin(self):
        """
        Test whether user is authenticated as the admin.
        """
        return self.authenticated and self.client['id'] == 'admin'

    def assert_has_permission(self, scope_required):
        """
        Warn that the required scope is not found in the scopes
        granted to the currently authenticated user.

        ::

            # The admin user should have client admin permissions
            uaa.assert_has_permission('admin', 'clients.admin')

        """
        if not self.authenticated:
            raise ValueError("Must first authenticate()")

        if scope_required not in self.get_scopes():
            logging.warning("Authenticated as %s" % (self.client['id']))
            logging.warning("Have scopes: %s" % (str.join(',', self.get_scopes())))
            logging.warning("Insufficient scope %s for operation" % (scope_required))

            raise ValueError("Client does not have permission.")

        return True

    def grant_client_permissions(self, client_id, admin=False, write=False,
            read=False, secret=False):
        """
        Grant the given client_id permissions for managing clients.

        - clients.admin: super user scope to create, modify, delete
        - clients.write: scope ot create and modify clients
        - clients.read: scope to read info about clients
        - clients.secret: scope to change password of a client

        """
        self.assert_has_permission('clients.admin')

        perms = []
        if admin:
            perms.append('clients.admin')

        if write or admin:
            perms.append('clients.write')

        if read or admin:
            perms.append('clients.read')

        if secret or admin:
            perms.append('clients.secret')

        if perms:
            self.update_client_grants(client_id, scope=perms,
                    authorities=perms)

    def get_clients(self):
        """
        Returns the clients stored in the instance of UAA.
        """
        self.assert_has_permission('clients.read')

        uri = self.uri + '/oauth/clients'
        headers = self.get_authorization_headers()
        response = requests.get(uri, headers=headers)
        return response.json()['resources']

    def get_client(self, client_id):
        """
        Returns details about a specific client by the client_id.
        """
        self.assert_has_permission('clients.read')

        uri = self.uri + '/oauth/clients/' + client_id
        headers = self.get_authorization_headers()
        response = requests.get(uri, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            # Not found but don't raise
            return

    def update_client_grants(self, client_id, scope=[], authorities=[],
            grant_types=[], redirect_uri=[], replace=False):
        """
        Will extend the client with additional scopes or
        authorities.  Any existing scopes and authorities will be left
        as is unless asked to replace entirely.
        """
        self.assert_has_permission('clients.write')

        client = self.get_client(client_id)
        if not client:
            raise ValueError("Must first create client: '%s'" % (client_id))

        if replace:
            changes = {
                'client_id': client_id,
                'scope': scope,
                'authorities': authorities,
                }
        else:
            changes = {'client_id': client_id}
            if scope:
                changes['scope'] = client['scope']
                changes['scope'].extend(scope)

            if authorities:
                changes['authorities'] = client['authorities']
                changes['authorities'].extend(authorities)

            if grant_types:
                if 'authorization_code' in grant_types and not redirect_uri:
                    logging.warning("A redirect_uri is required for authorization_code.")

                changes['authorized_grant_types'] = client['authorized_grant_types']
                changes['authorized_grant_types'].extend(grant_types)

            if redirect_uri:
                if 'redirect_uri' in client:
                    changes['redirect_uri'] = client['redirect_uri']
                    changes['redirect_uri'].extend(redirect_uri)
                else:
                    changes['redirect_uri'] = redirect_uri

        uri = self.uri + '/oauth/clients/' + client_id
        headers = {
            "pragma": "no-cache",
            "Cache-Control": "no-cache",
            "Content-Type": "application/json",
            "Accepts": "application/json",
            "Authorization": "Bearer " + self.get_token()
        }

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))
        logging.debug("BODY=" + json.dumps(changes))

        response = requests.put(uri, headers=headers, data=json.dumps(changes))

        logging.debug("STATUS=" + str(response.status_code))
        if response.status_code == 200:
            return response
        else:
            logging.error(response.content)
            response.raise_for_status()

    def grant_scim_permissions(self, client_id, read=False, write=False,
            create=False, userids=False, zones=False, invite=False,
            openid=False):
        """
        Grant the given client_id permissions for managing users.  System
        for Cross-domain Identity Management (SCIM) are required for accessing
        /Users and /Groups endpoints of UAA.

        - scim.read: scope for read access to all SCIM endpoints
        - scim.write: scope for write access to all SCIM endpoints
        - scim.create: scope to create/invite users and verify an account only
        - scim.userids: scope for id and username+origin conversion
        - scim.zones: scope for group management of users only
        - scim.invite: scope to participate in invitations
        - openid: scope to access /userinfo

        """
        self.assert_has_permission('clients.admin')

        perms = []
        if read:
            perms.append('scim.read')

        if write:
            perms.append('scim.write')

        if create:
            perms.append('scim.create')

        if userids:
            perms.append('scim.userids')

        if zones:
            perms.append('scim.zones')

        if invite:
            perms.append('scim.invite')

        if openid:
            perms.append('openid')

        if perms:
            self.update_client_grants(client_id, scope=perms, authorities=perms)

    def create_client(self, client_id, client_secret, manifest=None,
            client_credentials=True, refresh_token=True,
            authorization_code=False, redirect_uri=[]):
        """
        Will create a new client for your application use.

        - client_credentials: allows client to get access token
        - refresh_token: can be used to get new access token when expired
          without re-authenticating
        - authorization_code: redirection-based flow for user authentication

        More details about Grant types:
        - https://github.com/cloudfoundry/uaa/blob/master/docs/UAA-Security.md
        - https://tools.ietf.org/html/rfc6749

        A redirect_uri is required when using authorization_code.  See:
        https://www.predix.io/support/article/KB0013026

        """
        self.assert_has_permission('clients.admin')

        if authorization_code and not redirect_uri:
            raise ValueError("Must provide a redirect_uri for clients used with authorization_code")

        # Check if client already exists
        client = self.get_client(client_id)
        if client:
            return client

        uri = self.uri + '/oauth/clients'
        headers = {
            "pragma": "no-cache",
            "Cache-Control": "no-cache",
            "Content-Type": "application/json",
            "Accepts": "application/json",
            "Authorization": "Bearer " + self.get_token()
        }

        grant_types = []

        if client_credentials:
            grant_types.append('client_credentials')
        if refresh_token:
            grant_types.append('refresh_token')
        if authorization_code:
            grant_types.append('authorization_code')

        params = {
            "client_id": client_id,
            "client_secret": client_secret,
            "scope": ["uaa.none"],
            "authorized_grant_types": grant_types,
            "authorities": ["uaa.none"],
            "autoapprove": []
        }

        if redirect_uri:
            params.append(redirect_uri)

        response = requests.post(uri, headers=headers, data=json.dumps(params))
        if response.status_code == 201:
            if manifest:
                self.add_client_to_manifest(client_id, client_secret, manifest)

            client = {
                'id': client_id,
                'secret': client_secret
                }
            self._write_to_uaa_cache(client)
            return response
        else:
            logging.error(response.content)
            response.raise_for_status()

    def add_client_to_manifest(self, client_id, client_secret, manifest):
        """
        Add the given client / secret to the manifest for use in
        the application.
        """
        client_id_key = 'PREDIX_APP_CLIENT_ID'
        manifest.add_env_var(client_id_key, client_id)

        client_secret_key = 'PREDIX_APP_CLIENT_SECRET'
        manifest.add_env_var(client_secret_key, client_secret)

        manifest.write_manifest()

    def _post_user(self, data):
        """
        More complete control over creating users by calling the POST /Users
        directly.  The data should meet the UAA creation spec.

        More details:
        https://docs.cloudfoundry.org/api/uaa/#create-3
        """
        return self._post(self.uri + '/Users', data)

    def create_user(self, username, password, family_name, given_name, primary_email,
            details={}):
        """
        Creates a new user account with the required details.

        ::

            create_user('j12y', 'my-secret', 'Delancey', 'Jayson', 'volcano@ge.com')

        """
        self.assert_has_permission('scim.write')

        data = {
            'userName': username,
            'password': password,
            'name': {
                'familyName': family_name,
                'givenName': given_name,
                },
            'emails': [{
                'value': primary_email,
                'primary': True,
                }]
            }

        if details:
            data.update(details)

        return self._post_user(data)

    def delete_user(self, id):
        """
        Delete user with given id.
        """
        self.assert_has_permission('scim.write')

        uri = self.uri + '/Users/%s' % id
        headers = self._get_headers()

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))

        response = self.session.delete(uri, headers=headers)
        logging.debug("STATUS=" + str(response.status_code))
        if response.status_code == 200:
            return response
        else:
            logging.error(response.content)
            response.raise_for_status()

    def get_users(self, filter=None, sortBy=None, sortOrder=None,
            startIndex=None, count=None):
        """
        Returns users accounts stored in UAA.
        See https://docs.cloudfoundry.org/api/uaa/#list63

        For filtering help, see:
        http://www.simplecloud.info/specs/draft-scim-api-01.html#query-resources
        """
        self.assert_has_permission('scim.read')

        params = {}
        if filter:
            params['filter'] = filter

        if sortBy:
            params['sortBy'] = sortBy

        if sortOrder:
            params['sortOrder'] = sortOrder

        if startIndex:
            params['startIndex'] = startIndex

        if count:
            params['count'] = count

        return self._get(self.uri + '/Users', params=params)

    def get_user_by_username(self, username):
        """
        Returns details for user of the given username.

        If there is more than one match will only return the first.  Use
        get_users() for full result set.
        """
        results = self.get_users(filter='username eq "%s"' % (username))
        if results['totalResults'] == 0:
            logging.warning("Found no matches for given username.")
            return
        elif results['totalResults'] > 1:
            logging.warning("Found %s matches for username %s" %
                (results['totalResults'], username))

        return results['resources'][0]

    def get_user_by_email(self, email):
        """
        Returns details for user with the given email address.

        If there is more than one match will only return the first.  Use
        get_users() for full result set.
        """
        results = self.get_users(filter='email eq "%s"' % (email))
        if results['totalResults'] == 0:
            logging.warning("Found no matches for given email.")
            return
        elif results['totalResults'] > 1:
            logging.warning("Found %s matches for email %s" %
                (results['totalResults'], email))

        return results['resources'][0]

    def get_user(self, id):
        """
        Returns details about the user for the given id.

        Use get_user_by_email() or get_user_by_username() for help
        identifiying the id.
        """
        self.assert_has_permission('scim.read')
        return self._get(self.uri + '/Users/%s' % (id))

    def get_userinfo(self):
        """
        Retrieve user info for currently authenticated user.
        """
        return self._get(self.uri + '/userinfo')




import os
import logging

import predix.service


class ParkingPlanning(object):
    """
    The Parking Planning service provides (in beta) provides simulated data
    collected from street lights installed in specific cities.
    """

    # Constants for a few known bounding-box locations that have good data
    BBOX = {
            'USA': {
                'CA': {
                    'San Diego': '32.715675:-117.161230,32.708498:-117.151681',
                    }
                }
            }

    # Constants for device types
    DEVICE_TYPES = [
            'DATASIM'
            ]

    # Constants for codes representing asset types
    ASSET_TYPES = [
            'CAMERA',
            'ENV_SENSOR',
            'MIC',
            'NODE',
            'OTHERS'
            ]

    # Constants for codes represent event types
    EVENT_TYPES = [
            'HUMIDITY',
            'PEDEVT',
            'PKIN',
            'PKOUT',
            'PRESSURE',
            'TEMPERATURE',
            'TFEVT'
            ]

    # Constants for codes representing media
    MEDIA_TYPES = [
            'AUDIO',
            'IMAGE',
            'VIDEO',
            'OTHERS'
            ]

    def __init__(self):
        ns = 'predix.admin.ie.parking'
        self.zone_id = os.environ.get(ns + '.zone_id')
        if not self.zone_id:
            raise ValueError("%s.zone_id environment unset" % ns)

        self.uri = os.environ.get(ns + '.uri')
        if not self.uri:
            raise ValueError("%s.uri environment unset" % ns)

        self.service = predix.service.Service(self.zone_id)

    def authenticate_as_client(self, client_id, client_secret):
        """
        Will authenticate for the given client / secret.
        """
        self.service.uaa.authenticate(client_id, client_secret)

    def _get_headers(self):
        """
        Returns headers needed for making any service calls.
        """
        # Service will return a 500 if Accepts/Content-Type headers are used.
        headers = {
            'Predix-Zone-Id': self.zone_id,
            'Authorization': self.service._get_bearer_token()
            }
        return headers

    def _get_assets(self, bbox, size=None, page=None, asset_type=None,
            device_type=None, event_type=None, media_type=None):
        """
        Returns the raw results of an asset search for a given bounding
        box.
        """
        uri = self.uri + '/v1/assets/search'
        headers = self._get_headers()

        params = {
                'bbox': bbox,
                }

        # Query parameters

        params['q'] = []
        if device_type:
            if isinstance(device_type, str):
                device_type = [device_type]

            for device in device_type:
                if device not in self.DEVICE_TYPES:
                    logging.warning("Invalid device type: %s" % device)

                params['q'].append("device-type:%s" % device)

        if asset_type:
            if isinstance(asset_type, str):
                asset_type = [asset_type]

            for asset in asset_type:
                if asset not in self.ASSET_TYPES:
                    logging.warning("Invalid asset type: %s" % asset)
                params['q'].append("assetType:%s" % asset)

        if media_type:
            if isinstance(media_type, str):
                media_type = [media_type]

            for media in media_type:
                if media not in self.MEDIA_TYPES:
                    logging.warning("Invalid media type: %s" % media)
                params['q'].append("mediaType:%s" % media)

        if event_type:
            if isinstance(event_type, str):
                event_type = [event_type]

            for event in event_type:
                if event not in self.EVENT_TYPES:
                    logging.warning("Invalid event type: %s" % event)
                params['q'].append("eventTypes:%s" % event)

        # Pagination parameters

        if size:
            params['size'] = size

        if page:
            params['page'] = page

        return self.service._get(uri, params=params, headers=headers)

    def get_assets(self, bbox, **kwargs):
        """
        Query the assets stored in the intelligent environment for a given
        bounding box and query.

        Assets can be filtered by type of asset, event, or media available.

            - device_type=['DATASIM']
            - asset_type=['CAMERA']
            - event_type=['PKIN']
            - media_type=['IMAGE']

        Pagination can be controlled with keyword parameters

            - page=2
            - size=100

        Returns a list of assets stored in a dictionary that describe their:

            - asset-id
            - device-type
            - device-id
            - media-type
            - coordinates
            - event-type

        Additionally there are some _links for additional information.
        """
        response = self._get_assets(bbox, **kwargs)

        # Remove broken HATEOAS _links but identify asset uid first
        assets = []
        for asset in response['_embedded']['assets']:
            asset_url = asset['_links']['self']
            uid = asset_url['href'].split('/')[-1]
            asset['uid'] = uid

            del(asset['_links'])
            assets.append(asset)

        return assets

    def _get_asset(self, asset_uid):
        """
        Returns raw response for an given asset by its unique id.
        """
        uri = self.uri + '/v2/assets/' + asset_uid

        headers = self._get_headers()

        return self.service._get(uri, headers=headers)

    def get_asset(self, asset_uid):
        """
        Return details for a given asset.
        """
        response = self._get_asset(asset_uid)

        del(response['_links'])

        return response





import os
import re
import time
import json
import logging
import datetime
import websocket

import predix.config
import predix.service


class TimeSeries(object):
    """
    Client library for working with the Time Series service.

    :param query_uri: URI for Time Series endpoint to execute queries, can be
        derived from environment PREDIX_TIMESERIES_QUERY_URI variable if not
        passed.

    :param ingest_uri: URI for Time Series endpoint to ingest data, can be
        derived from environment PREDIX_TIMESERIES_INGEST_URI variable if not
        passed.

    :param query_zone_id: Predix-Zone-Id for Query Authorization, can be
        derived from environment PREDIX_TIMESERIES_QUERY_ZONE_ID.

    :param ingest_zone_id: Predix-Zone-Id for Ingest Authorization, can be
        derived from environment PREDIX_TIMESERIES_INGEST_ZONE_ID.

    :param read: Whether we expect to be able to query / read data.

    :param write: Whether we expect to be able to ingest / write data.

    Learn more about Predix Time Series:
    https://www.predix.io/services/service.html?id=1177

    """
    # "Constants" for data quality values
    BAD = 0
    UNCERTAIN = 1
    NA = 2
    GOOD = 3

    def __init__(self, read=True, write=True, query_uri=None, ingest_uri=None,
            query_zone_id=None, ingest_zone_id=None, *args, **kwargs):
        """
        Time Series by default will grant the client both read
        and write permissions.  Either can be disabled.
        """
        super(TimeSeries, self).__init__(*args, **kwargs)

        # Not all clients can read and write, so look to arguments on whether
        # to validate permissions during initialization.
        if read:
            self.query_uri = query_uri or self._get_query_uri()
            self.query_zone_id = query_zone_id or self._get_query_zone_id()

        if write:
            self.ingest_uri = ingest_uri or self._get_ingest_uri()
            self.ingest_zone_id = ingest_zone_id or self._get_ingest_zone_id()

        self.zone_id = self.query_zone_id or self.ingest_zone_id
        self.service = predix.service.Service(self.zone_id)

        # Store a websocket connection once opened
        self.ws = None

        # Store in-memory and forward any datapoints as a single transaction
        self._queue = []

    def _get_query_uri(self):
        """
        Returns the URI endpoint for performing queries of a
        Predix Time Series instance from environment inspection.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_timeseries = services['predix-timeseries'][0]['credentials']
            return predix_timeseries['query']['uri'].partition('/v1')[0]
        else:
            return predix.config.get_env_value(self, 'query_uri')

    def _get_query_zone_id(self):
        """
        Returns the ZoneId for performing queries of a Predix
        Time Series instance from environment inspection.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_timeseries = services['predix-timeseries'][0]['credentials']
            return predix_timeseries['query']['zone-http-header-value']
        else:
            return predix.config.get_env_value(self, 'query_zone_id')

    def _get_ingest_uri(self):
        """
        Returns the URI endpoint for performing ingestion of data to
        Predix Time Series instance from environment inspection.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_timeseries = services['predix-timeseries'][0]['credentials']
            return predix_timeseries['ingest']['uri']
        else:
            return predix.config.get_env_value(self, 'ingest_uri')

    def _get_ingest_zone_id(self):
        """
        Returns the ZoneId for ingesting data to a Predix
        Time Series instance from environment inspection.
        """
        if 'VCAP_SERVICES' in os.environ:
            services = json.loads(os.getenv('VCAP_SERVICES'))
            predix_timeseries = services['predix-timeseries'][0]['credentials']
            return predix_timeseries['ingest']['zone-http-header-value']
        else:
            return predix.config.get_env_value(self, 'ingest_zone_id')

    def __del__(self):
        """
        Destructor to make sure an open websocket connection is closed.
        """
        # No need to delete if not properly initialized
        if not hasattr(self, '_queue'):
            return

        if len(self._queue) > 0:
            logging.warning("%s buffered datapoints in queue lost." %
                    (len(self._queue)))

        if self.ws:
            self.ws.close()

    def authenticate_as_client(self, client_id, client_secret):
        """
        Will authenticate for the given client / secret.
        """
        self.service.uaa.authenticate(client_id, client_secret)

    def _get_aggregations(self):
        """
        Returns all of the aggregations in time series.  There is no support
        in the service for filtering or paginations.
        """
        url = self.query_uri + '/v1/aggregations'
        return self.service._get(url)

    def get_aggregations(self):
        """
        Returns all of the aggregations that can be used with the
        Time Series Service.
        """
        return self._get_aggregations()['results']

    def _get_tags(self):
        """
        Returns all of the tags in time series.  There is no support
        in the service for filtering or paginations.
        """
        return self.service._get(self.query_uri + '/v1/tags')

    def get_tags(self):
        """
        Returns a list of the tag names in the timeseries service
        instance.
        """
        return self._get_tags()['results']

    def _get_datapoints(self, params):
        """
        Will make a direct REST call with the given json body payload to
        get datapoints.
        """
        url = self.query_uri + '/v1/datapoints'
        return self.service._get(url, params=params)

    def _post_datapoints(self, body):
        url = self.query_uri + '/v1/datapoints'
        return self.service._post(url, body)

    def get_values(self, *args, **kwargs):
        """
        Convenience method that for simple single tag queries will
        return just the values to be iterated on.
        """
        if isinstance(args[0], list):
            raise ValueError("Can only get_values() for a single tag.")

        response = self.get_datapoints(*args, **kwargs)
        for value in response['tags'][0]['results'][0]['values']:
            yield [datetime.datetime.utcfromtimestamp(value[0]/1000),
                   value[1],
                   value[2]]

    def is_good(self, quality):
        """
        Simple test if data quality is GOOD.
        """
        return quality == ts.GOOD

    def is_bad(self, quality):
        """
        Simple test if data quality value is BAD.
        """
        return quality == ts.BAD

    def get_datapoints(self, tags, start=None, end=None, order=None,
            limit=None, qualities=None, attributes=None, measurement=None,
            aggregations=None, post=False):
        """
        Returns all of the datapoints that match the given query.

            - tags: list or string identifying the name/tag (ie. "temp")
            - start: data after this, absolute or relative (ie. '1w-ago' or
              1494015972386)
            - end: data before this value
            - order: ascending (asc) or descending (desc)
            - limit: only return a few values (ie. 25)
            - qualities: data quality value (ie. [ts.GOOD, ts.UNCERTAIN])
            - attributes: dictionary of key-values (ie. {'unit': 'mph'})
            - measurement: tuple of operation and value (ie. ('gt', 30))
            - aggregations: summary statistics on data results (ie. 'avg')
            - post: POST query instead of GET (caching implication)

        A few additional observations:
            - allow service to do most data validation
            - order is applied before limit so resultset will differ

        The returned results match what the service response is so you'll
        need to unpack it as appropriate.  Oftentimes what you want for
        a simple single tag query will be:

            response['tags'][0]['results'][0]['values']

        """
        params = {}

        # Documentation says start is required for GET but not POST, but
        # seems to be required all the time, so using sensible default.
        if not start:
            start = '1w-ago'
            logging.warning("Defaulting query for data with start date %s" % (start))

        # Start date can be absolute or relative, only certain legal values
        # but service will throw error if used improperly.  (ms, s, mi, h, d,
        # w, mm, y).  Relative dates must end in -ago.
        params['start'] = start

        # Docs say when making POST with a start that end must also be
        # specified, but this does not seem to be the case.
        if end:
            # MAINT: error when end < start which is handled by service
            params['end'] = end

        params['tags'] = []
        if not isinstance(tags, list):
            tags = [tags]

        for tag in tags:
            query = {}
            query['name'] = tag

            # Limit resultset with an integer value
            if limit:
                query['limit'] = int(limit)

            # Order must be 'asc' or 'desc' but will get sensible error
            # from service.
            if order:
                query['order'] = order

            # Filters are complex and support filtering by
            # quality, measurement, and attributes.
            filters = {}

            # Check for the quality of the datapoints
            if qualities is not None:
                if isinstance(qualities, int) or isinstance(qualities, str):
                    qualities = [qualities]

                # Timeseries expects quality to be a string, not integer,
                # so coerce each into a string
                for i, quality in enumerate(qualities):
                    qualities[i] = str(quality)

                filters['qualities'] = {"values": qualities}

            # Check for attributes on the datapoints, expected to be
            # a dictionary of key / value pairs that datapoints must match.
            if attributes is not None:
                if not isinstance(attributes, dict):
                    raise ValueError("Attribute filters must be dictionary.")

                filters['attributes'] = attributes

            # Check for measurements that meets a given comparison operation
            # such as ge, gt, eq, ne, le, lt
            if measurement is not None:
                filters['measurements'] = {
                        'condition': measurement[0],
                        'values': measurement[1]
                        }

            # If we found any filters add them to the query
            if filters:
                query['filters'] = filters

            # Handle any additional aggregations of dataset
            if aggregations is not None:
                if not isinstance(aggregations, list):
                    aggregations = [aggregations]

                query['aggregations'] = []
                for aggregation in aggregations:
                    query['aggregations'].append({
                        'sampling': {'datapoints': 1},
                        'type': aggregation })

            params['tags'].append(query)

        if post:
            return self._post_datapoints(params)
        else:
            return self._get_datapoints({"query": json.dumps(params)})

    def _get_latest(self, params):
        """
        Will directly pass the given params as parameters
        to the GET endpoint.
        """
        uri = self.query_uri + '/v1/datapoints/latest'
        return self.service._get(uri, params=params)

    def _post_latest(self, body):
        """
        Will directly pass the given body object as the JSON
        payload to a POST call.
        """
        uri = self.query_uri + '/v1/datapoints/latest'
        return self.service._post(uri, body)

    def get_latest(self, tags, post=False):
        """
        Similar to get_datapoints() but will only return the very
        last datapoint ingested for the given tag or tags (if a list).

        If given the optional post parameter will make the call as a POST
        rather than a GET request.  The GET can have an advantage of url
        based response caching and the POST has the advantage of a larger
        and more complex query.
        """
        params = {}
        if isinstance(tags, list):
            params['tags'] = str.join(',', tags)
        elif isinstance(tags, str):
            params['tags'] = tags
        else:
            raise ValueError("Expect to get tags as a str or list")

        if post:
            return self._post_latest(params)
        else:
            return self._get_latest(params)

    def _create_connection(self):
        """
        Create a new websocket connection with proper headers.
        """
        logging.debug("Initializing new websocket connection.")
        headers = {
            'Authorization': self.service._get_bearer_token(),
            'Predix-Zone-Id': self.ingest_zone_id,
            'Content-Type': 'application/json',
        }
        url = self.ingest_uri

        logging.debug("URL=" + str(url))
        logging.debug("HEADERS=" + str(headers))

        # Should consider connection pooling and longer timeouts
        return websocket.create_connection(url, header=headers)

    def _get_websocket(self, reuse=True):
        """
        Reuse existing connection or create a new connection.
        """
        # Check if still connected
        if self.ws and reuse:
            if self.ws.connected:
                return self.ws

            logging.debug("Stale connection, reconnecting.")

        self.ws = self._create_connection()
        return self.ws

    def _send_to_timeseries(self, message):
        """
        Establish or reuse socket connection and send
        the given message to the timeseries service.
        """
        logging.debug("MESSAGE=" + str(message))

        result = None
        try:
            ws = self._get_websocket()
            ws.send(json.dumps(message))
            result = ws.recv()
        except (websocket.WebSocketConnectionClosedException, Exception) as e:
            logging.debug("Connection failed, will try again.")
            logging.debug(e)

            ws = self._get_websocket(reuse=False)
            ws.send(json.dumps(message))
            result = ws.recv()

        logging.debug("RESULT=" + str(result))
        return result

    def queue(self, name, value, quality=None, timestamp=None,
            attributes=None):
        """
        To reduce network traffic, you can buffer datapoints and
        then flush() anything in the queue.

        :param name: the name / label / tag for sensor data

        :param value: the sensor reading or value to record

        :param quality: the quality value, use the constants BAD, GOOD, etc.
            (optional and defaults to UNCERTAIN)

        :param timestamp: the time the reading was recorded in epoch
            milliseconds (optional and defaults to now)

        :param attributes: dictionary for any key-value pairs to store with the
            reading (optional)

        """
        # Get timestamp first in case delay opening websocket connection
        # and it must have millisecond accuracy
        if not timestamp:
            timestamp = int(round(time.time() * 1000))
        else:
            # Coerce datetime objects to epoch
            if isinstance(timestamp, datetime.datetime):
                timestamp = int(round(int(timestamp.strftime('%s')) * 1000))

        # Only specific quality values supported
        if quality not in [self.BAD, self.GOOD, self.NA, self.UNCERTAIN]:
            quality = self.UNCERTAIN

        # Check if adding to queue of an existing tag and add second datapoint
        for point in self._queue:
            if point['name'] == name:
                point['datapoints'].append([timestamp, value, quality])
                return

        # If adding new tag, initialize and set any attributes
        datapoint = {
            "name": name,
            "datapoints": [[timestamp, value, quality]]
        }

        # Attributes are extra details for a datapoint

        if attributes is not None:
            if not isinstance(attributes, dict):
                raise ValueError("Attributes are expected to be a dictionary.")

            # Validate rules for attribute keys to provide guidance.
            invalid_value = ':;= '
            has_invalid_value = re.compile(r'[%s]' % (invalid_value)).search
            has_valid_key = re.compile(r'^[\w\.\/\-]+$').search

            for (key, val) in list(attributes.items()):
                # Values cannot be empty
                if (val == '') or (val is None):
                    raise ValueError("Attribute (%s) must have a non-empty value." % (key))

                # Values should be treated as a string for regex validation
                val = str(val)

                # Values cannot contain certain arbitrary characters
                if bool(has_invalid_value(val)):
                    raise ValueError("Attribute (%s) cannot contain (%s)." %
                            (key, invalid_value))

                # Attributes have to be alphanumeric-ish
                if not bool(has_valid_key):
                    raise ValueError("Key (%s) not alphanumeric-ish." % (key))

            datapoint['attributes'] = attributes

        self._queue.append(datapoint)
        logging.debug("QUEUE: " + str(len(self._queue)))

    def send(self, name=None, value=None, **kwargs):
        """
        Can accept a name/tag and value to be queued and then send anything in
        the queue to the time series service.  Optional parameters include
        setting quality, timestamp, or attributes.

        See spec for queue() for complete list of options.

        Example of sending a batch of values:

            queue('temp', 70.1)
            queue('humidity', 20.4)
            send()

        Example of sending one and flushing queue immediately

            send('temp', 70.3)
            send('temp', 70.4, quality=ts.GOOD, attributes={'unit': 'F'})


        """
        if name and value:
            self.queue(name, value, **kwargs)

        timestamp = int(round(time.time() * 1000))

        # The label "name" or "tag" is sometimes used ambiguously
        msg = {
            "messageId": timestamp,
            "body": self._queue
        }

        self._queue = []

        return self._send_to_timeseries(msg)



import json
import logging
import threading
import websocket

import time

from predix.data.eventhub import EventHub_pb2, EventHub_pb2_grpc
from predix.data.eventhub.client import Eventhub


class PublisherConfig:
    """
    object to store the publisher config
    as well as some of the enums for the settings
    """

    class AcknowledgementOptions:
        def __init__(self):
            pass

        NACKS_ONLY = 'NACKS_ONLY'
        ACKS_AND_NACKS = 'ACKS_AND_NACKS'
        NONE = 'NONE'

    class Protocol:
        def __init__(self):
            pass

        WSS = 'WSS'
        GRPC = 'GRPC'

    class Type:
        def __init__(self):
            pass

        ASYNC = 'ASYNC'
        SYNC = 'SYNC'

    def __init__(self,
                 topic="",
                 publish_type=Type.ASYNC,
                 sync_timeout=15,
                 protocol=Protocol.GRPC,
                 async_cache_ack_interval_millis=500,
                 async_cache_acks_and_nacks=True,
                 async_acknowledgement_options=AcknowledgementOptions.ACKS_AND_NACKS,
                 async_auto_send=False,
                 async_auto_send_amount=100,
                 async_auto_send_interval_millis=10000):
        """

        :param topic: str the topic to publish to
        :param publish_type: the type of publisher you want to make. sync means publish is a blocking call
        :param sync_timeout: timeout length to wait for acks
        :param protocol: the protocol for publishing messages, WSS or GRPC
        :param async_cache_ack_interval_millis:
        :param async_cache_acks_and_nacks: should service cache the acks
        :param async_acknowledgement_options: what acks should be received
        :param async_auto_send: should the skd auto send messages
        :param async_auto_send_amount: after how many messages should messages be automatically sent
        :param async_auto_send_interval_millis: what is the max period between messages
        """

        self.topic = topic
        self.protocol = protocol
        self.publish_type = publish_type

        # Async config options
        self.async_cache_ack_interval_millis = async_cache_ack_interval_millis
        self.async_cache_acks_and_nacks = async_cache_acks_and_nacks
        if async_acknowledgement_options == self.AcknowledgementOptions.ACKS_AND_NACKS:
            self.async_enable_acks = True
            self.async_enable_nacks_only = False
        elif async_acknowledgement_options == self.AcknowledgementOptions.NACKS_ONLY:
            self.enable_acks = False
            self.enable_nacks_only = True
        elif async_acknowledgement_options == self.AcknowledgementOptions.NONE:
            self.async_enable_acks = False
            self.async_enable_nacks_only = False
        self.async_auto_send = async_auto_send
        self.async_auto_send_amount = async_auto_send_amount
        self.async_auto_send_interval_millis = async_auto_send_interval_millis

        # sync config options
        self.sync_timeout = sync_timeout

    def is_grpc(self):
        return self.protocol == self.Protocol.GRPC

    def is_wss(self):
        return self.protocol == self.Protocol.WSS

    def is_async(self):
        return self.publish_type == self.Type.ASYNC

    def is_sync(self):
        return self.publish_type == self.Type.SYNC


class Publisher:
    """
    Publisher Object for both grpc and web socket
    """

    def __init__(self, eventhub_client, config, channel=None):
        self.eventhub_client = eventhub_client
        self._channel = channel
        self.config = config
        self._ws = None
        if config.is_wss():
            self._init_publisher_ws()
        else:
            if channel is None:
                raise ValueError("must provide channel if using grpc to publish")
            self._init_grpc_publisher()

        # Operational
        self._rx_queue = []
        self._tx_queue = []
        self._tx_queue_lock = threading.Lock()
        self.callback = None
        self.last_send_time = 0
        self._run_ack_generator = True
        self._active = True

        if config.async_auto_send:
            t = threading.Thread(target=self._auto_send)
            t.daemon = True
            t.start()

    def __del__(self):
        """
        Destructor to make sure an open web socket connection is closed.
        """
        if self._active:
            if self._ws is not None:
                logging.debug("closing websocket")
                self._ws.close()
                logging.debug("waiting on websocket thread to exit")
                self._ws_thread.join()
            else:
                logging.debug("stopping generators")
                self.grpc_manager.stop_generator()
                self._run_ack_generator = False

    """
    ####################################################################################
    User Code
    ####################################################################################
    """

    def shutdown(self):
        """
        shutdown the client, stops the web socket or grpc threads
        :return:
        """
        if self._active:
            if self._ws is not None:
                logging.debug("closing web socket")
                self._ws.close()
                logging.debug("waiting on webs ocket thread to exit")
                self._ws_thread.join()
                self._ws = None
            else:
                logging.debug("stopping generators")
                self.grpc_manager.stop_generator()
                self._run_ack_generator = False
        self._active = False

    def add_message(self, id, body, tags=False):
        """
        add messages to the rx_queue
        :param id: str message Id
        :param body: str the message body
        :param tags: dict[string->string] tags to be associated with the message
        :return: self
        """
        if not tags:
            tags = {}
        try:
            self._tx_queue_lock.acquire()
            self._tx_queue.append(
                EventHub_pb2.Message(id=id, body=body, tags=tags, zone_id=self.eventhub_client.zone_id))
        finally:
            self._tx_queue_lock.release()
        return self

    def publish_queue(self):
        """
        Publish all messages that have been added to the queue for configured protocol
        :return: None
        """
        self.last_send_time = time.time()
        try:
            self._tx_queue_lock.acquire()
            start_length = len(self._rx_queue)
            publish_amount = len(self._tx_queue)
            if self.config.protocol == PublisherConfig.Protocol.GRPC:
                self._publish_queue_grpc()
            else:
                self._publish_queue_wss()
            self._tx_queue = []
        finally:
            self._tx_queue_lock.release()

        if self.config.publish_type == self.config.Type.SYNC:
            start_time = time.time()
            while time.time() - start_time < self.config.sync_timeout and \
                                    len(self._rx_queue) - start_length < publish_amount:
                pass
            return self._rx_queue

    def ack_generator(self):
        """
        generator for acks to yield messages to the user in a async configuration
        :return: messages as they come in
        """
        if self.config.is_sync():
            logging.warning('cant use generator on a sync publisher')
            return
        while self._run_ack_generator:
            while len(self._rx_queue) != 0:
                logging.debug('yielding to client')
                yield self._rx_queue.pop(0)
        return

    """
    ####################################################################################
    Internal
    ####################################################################################
    """

    def _auto_send(self):
        """
        auto send blocking function, when the interval or the message size has been reached, publish
        :return:
        """
        while True:
            if time.time() - self.last_send_time > self.config.async_auto_send_interval_millis or \
                            len(self._tx_queue) >= self.config.async_auto_send_amount:
                self.publish_queue()

    def _generate_publish_headers(self):
        """
        generate the headers for the connection to event hub service based on the provided config
        :return: {} headers
        """
        headers = {
            'predix-zone-id': self.eventhub_client.zone_id
        }
        token = self.eventhub_client.service._get_bearer_token()
        if self.config.is_grpc():
            headers['authorization'] = token[(token.index(' ') + 1):]
        else:
            headers['authorization'] = token

        if self.config.topic == '':
            headers['topic'] = self.eventhub_client.zone_id + '_topic'
        else:
            headers['topic'] = self.config.topic

        if self.config.publish_type == self.config.Type.SYNC:
            headers['sync-acks'] = 'true'
        else:
            headers['sync-acks'] = 'false'
            headers['send-acks-interval'] = str(self.config.async_cache_ack_interval_millis)
            headers['acks'] = str(self.config.async_enable_acks).lower()
            headers['nacks'] = str(self.config.async_enable_nacks_only).lower()
            headers['cache-acks'] = str(self.config.async_cache_acks_and_nacks).lower()
        return headers

    def _publisher_callback(self, publish_ack):
        """
        publisher callback that grpc and web socket can pass messages to
        address the received message onto the queue
        :param publish_ack: EventHub_pb2.Ack the ack received from either wss or grpc
        :return: None
        """
        logging.debug("ack received: " + str(publish_ack).replace('\n', ' '))
        self._rx_queue.append(publish_ack)

    """
    ####################################################################################
    GRPC  Code
    ####################################################################################
    """

    def _init_grpc_publisher(self):
        """
        initialize the grpc publisher, builds the stub and then starts the grpc manager
        :return: None
        """
        self._stub = EventHub_pb2_grpc.PublisherStub(channel=self._channel)
        self.grpc_manager = Eventhub.GrpcManager(stub_call=self._stub.send,
                                                 on_msg_callback=self._publisher_callback,
                                                 metadata=self._generate_publish_headers().items())

    def _publish_queue_grpc(self):
        """
        send the messages in the tx queue to the GRPC manager
        :return: None
        """
        messages = EventHub_pb2.Messages(msg=self._tx_queue)
        publish_request = EventHub_pb2.PublishRequest(messages=messages)
        self.grpc_manager.send_message(publish_request)

    """
    ####################################################################################
    Web Socket Code
    ####################################################################################
    """

    def _publish_queue_wss(self):
        """
        send the messages down the web socket connection as a json object
        :return: None
        """

        msg = []
        for m in self._tx_queue:
            msg.append({'id': m.id, 'body': m.body, 'zone_id': m.zone_id})
        self._ws.send(json.dumps(msg), opcode=websocket.ABNF.OPCODE_BINARY)

    def _init_publisher_ws(self):
        """
        Create a new web socket connection with proper headers.
        """
        logging.debug("Initializing new web socket connection.")

        url = ('wss://%s/v1/stream/messages/' % self.eventhub_client.host)

        headers = self._generate_publish_headers()

        logging.debug("URL=" + str(url))
        logging.debug("HEADERS=" + str(headers))

        websocket.enableTrace(False)
        self._ws = websocket.WebSocketApp(url,
                                          header=headers,
                                          on_message=self._on_ws_message,
                                          on_open=self._on_ws_open,
                                          on_close=self._on_ws_close)
        self._ws_thread = threading.Thread(target=self._ws.run_forever, kwargs={'ping_interval': 30})
        self._ws_thread.daemon = True
        self._ws_thread.start()
        time.sleep(1)

    def _on_ws_message(self, ws, message):
        """
        on_message callback of websocket class, load the message into a dict and then
        update an Ack Object with the results
        :param ws: web socket connection that the message was received on
        :param message: web socket message in text form
        :return: None
        """
        logging.debug(message)
        json_list = json.loads(message)
        for rx_ack in json_list:
            ack = EventHub_pb2.Ack()
            for key, value in rx_ack.items():
                setattr(ack, key, value)
            self._publisher_callback(ack)

    def _on_ws_open(self, ws):
        """
        on open callback of websocket connection
        :param ws: web socket connection that the message was received on
        :return: None
        """
        logging.debug("ws web socket connected")
        pass

    def _on_ws_error(self, ws, error):
        """
        on error callback of web socket error
        :param ws: web socket connection that the message was received on
        :param error:
        :return:
        """
        logging.debug(error)
        pass

    def _on_ws_close(self, ws):
        """
        on close callback of websocket class
        :param ws:  web socket connection that the message was received on
        :return: None
        """
        logging.debug("ws connection closed")
        pass




import predix.admin.cf.api
import logging

class Org(object):
    """
    Operations and data for Cloud Foundry Organizations.
    """
    ROLE_MAP = {
        'user': '/v2/organizations/%s/users',
        'auditor': '/v2/organizations/%s/auditors',
        'manager': '/v2/organizations/%s/managers',
        'billing_manager': '/v2/organizations/%s/billing_managers'
    }
    
    def __init__(self, *args, **kwargs):
        super(Org, self).__init__(*args, **kwargs)

        self.api = predix.admin.cf.api.API()

        self.name = self.api.config.get_organization_name()
        self.guid = self.api.config.get_organization_guid()

    def _get_orgs(self):
        """
        Returns the organizations for the authenticated user.
        """
        return self.api.get('/v2/organizations')

    def get_orgs(self):
        """
        Returns a flat list of the names for the organizations
        user belongs.
        """
        orgs = []
        for resource in self._get_orgs()['resources']:
            orgs.append(resource['entity']['name'])

        return orgs

    def _get_apps(self):
        """
        Returns all of the apps in the organization.
        """
        return self.api.get('/v2/apps')

    def get_apps(self):
        """
        Returns a flat list of the names for the apps in
        the organization.
        """
        apps = []
        for resource in self._get_apps()['resources']:
            apps.append(resource['entity']['name'])

        return apps

    def has_app(self, app_name):
        """
        Simple test to see if we have a name conflict
        for the application.
        """
        return app_name in self.get_apps()

    def get_users(self):
        return self.api.get('/v2/organizations/%s/users' % self.guid)

    def add_user(self, user_name, role='user'):
        """
        Calls CF's associate user with org. Valid roles include `user`, `auditor`,
        `manager`,`billing_manager`
        """
        role_uri = self._get_role_uri(role=role)
        return self.api.put(path=role_uri, data={'username': user_name})

    def _get_role_uri(self, role):
        try:
            role_uri = self.ROLE_MAP[role]
            return role_uri % self.api.config.get_organization_guid()
        except KeyError:
            logging.error('"%s" role is not a valid role' % role)
            raise

    def remove_user(self, user_name, role):
        """
        Calls CF's remove user with org
        """
        role_uri = self._get_role_uri(role=role)
        return self.api.delete(path=role_uri, data={'username': user_name})




import os
import time

import predix.config
import predix.admin.service
import predix.data.cache

class Cache(object):
    """
    Predix Cache key-value persistent storage.
    """
    def __init__(self, plan_name=None, name=None, uaa=None, *args, **kwargs):
        super(Cache, self).__init__(*args, **kwargs)
        self.service_name = 'predix-cache'
        self.plan_name = plan_name or 'Shared-R30'
        self.use_class = predix.data.cache.Cache

        self.service = predix.admin.service.CloudFoundryService(self.service_name,
                self.plan_name, name=name)

    def exists(self):
        """
        Returns whether or not this service already exists.
        """
        return self.service.exists()

    def _create_in_progress(self):
        """
        Creating this service is handled asynchronously so this method will
        simply check if the create is in progress.  If it is not in progress,
        we could probably infer it either failed or succeeded.
        """
        instance = self.service.service.get_instance(self.service.name)
        if (instance['last_operation']['state'] == 'in progress' and
           instance['last_operation']['type'] == 'create'):
               return True

        return False

    def create(self, max_wait=180, **kwargs):
        """
        Create an instance of the Predix Cache Service with they typical
        starting settings.

        :param max_wait: service is created asynchronously, so will only wait
            this number of seconds before giving up.

        """
        # Will need to wait for the service to be provisioned before can add
        # service keys and get env details.
        self.service.create(async=True, create_keys=False)
        while self._create_in_progress() and max_wait > 0:
            time.sleep(1)
            max_wait -= 1

        # Now get the service env (via service keys)
        cfg = self.service._get_service_config()
        self.service.settings.save(cfg)

        host = predix.config.get_env_key(self.use_class, 'host')
        os.environ[host] = self.service.settings.data['host']

        password = predix.config.get_env_key(self.use_class, 'password')
        os.environ[password] = self.service.settings.data['password']

        port = predix.config.get_env_key(self.use_class, 'port')
        os.environ[port] = str(self.service.settings.data['port'])

    def add_to_manifest(self, manifest):
        """
        Add useful details to the manifest about this service so
        that it can be used in an application.

        :param manifest: A predix.admin.app.Manifest object instance
            that manages reading/writing manifest config for a
            cloud foundry app.
        """
        manifest.add_service(self.service.name)

        host = predix.config.get_env_key(self.use_class, 'host')
        manifest.add_env_var(host, self.service.settings.data['host'])

        password = predix.config.get_env_key(self.use_class, 'password')
        manifest.add_env_var(password, self.service.settings.data['password'])

        port = predix.config.get_env_key(self.use_class, 'port')
        manifest.add_env_var(port, self.service.settings.data['port'])

        manifest.write_manifest()




import os
import time
import logging

import predix.config
import predix.admin.service
import predix.data.dbaas

class PostgreSQL(object):
    """
    Predix Database as a Service provides a relational database management system
    (PostgreSQL) for data persistence.
    """
    def __init__(self, plan_name=None, name=None, uaa=None, *args, **kwargs):
        super(PostgreSQL, self).__init__(*args, **kwargs)
        self.service_name = 'postgres-2.0'
        self.plan_name = plan_name or 'dedicated-1.1'
        self.use_class = predix.data.dbaas.PostgreSQL

        self.service = predix.admin.service.CloudFoundryService(self.service_name,
                self.plan_name, name=name)

    def exists(self):
        """
        Returns whether or not this service already exists.
        """
        return self.service.exists()

    def _create_in_progress(self):
        """
        Creating this service is handled asynchronously so this method will
        simply check if the create is in progress.  If it is not in progress,
        we could probably infer it either failed or succeeded.
        """
        instance = self.service.service.get_instance(self.service.name)
        if (instance['last_operation']['state'] == 'in progress' and
           instance['last_operation']['type'] == 'create'):
               return True

        return False

    def create(self, max_wait=300, allocated_storage=None,
            encryption_at_rest=None, restore_to_time=None, **kwargs):
        """
        Create an instance of the PostgreSQL service with the typical starting
        settings.

        :param max_wait: service is created asynchronously, so will only wait
            this number of seconds before giving up.

        :param allocated_storage: int for GBs to be allocated for storage

        :param encryption_at_rest: boolean for encrypting data that is stored

        :param restore_to_time: UTC date within recovery period for db
            backup to be used when initiating

        """

        # MAINT: Add these if there is demand for it and validated

        if allocated_storage or encryption_at_rest or restore_to_time:
            raise NotImplementedError()

        # Will need to wait for the service to be provisioned before can add
        # service keys and get env details.

        self.service.create(async=True, create_keys=False)
        while self._create_in_progress() and max_wait > 0:
            if max_wait % 5 == 0:
                logging.warning('Can take {}s for create to finish.'.format(max_wait))
            time.sleep(1)
            max_wait -= 1

        # Now get the service env (via service keys)
        cfg = self.service._get_service_config()
        self.service.settings.save(cfg)

        hostname = predix.config.get_env_key(self.use_class, 'hostname')
        os.environ[hostname] = self.service.settings.data['hostname']

        password = predix.config.get_env_key(self.use_class, 'password')
        os.environ[password] = self.service.settings.data['password']

        port = predix.config.get_env_key(self.use_class, 'port')
        os.environ[port] = str(self.service.settings.data['port'])

        username = predix.config.get_env_key(self.use_class, 'username')
        os.environ[username] = self.service.settings.data['username']

        uri = predix.config.get_env_key(self.use_class, 'uri')
        os.environ[uri] = self.service.settings.data['uri']

    def add_to_manifest(self, manifest):
        """
        Add useful details to the manifest about this service so
        that it can be used in an application.

        :param manifest: A predix.admin.app.Manifest object instance
            that manages reading/writing manifest config for a
            cloud foundry app.
        """
        manifest.add_service(self.service.name)

        hostname = predix.config.get_env_key(self.use_class, 'hostname')
        manifest.add_env_var(hostname, self.service.settings.data['hostname'])

        password = predix.config.get_env_key(self.use_class, 'password')
        manifest.add_env_var(password, self.service.settings.data['password'])

        port = predix.config.get_env_key(self.use_class, 'port')
        manifest.add_env_var(port, self.service.settings.data['port'])

        username = predix.config.get_env_key(self.use_class, 'username')
        manifest.add_env_var(username, self.service.settings.data['username'])

        uri = predix.config.get_env_key(self.use_class, 'uri')
        manifest.add_env_var(uri, self.service.settings.data['uri'])

        manifest.write_manifest()




import os
import logging

from cryptography.fernet import Fernet


def is_cf_env():
    # MAINT: consider CF_INSTANCE_ADDR
    return ('VCAP_SERVICES' in os.environ) or ('VCAP_APPLICATION' in os.environ)

def get_crypt_key(key_path):
    """
    Get the user's PredixPy manifest key.  Generate and store one if not
    yet generated.
    """
    key_path = os.path.expanduser(key_path)
    if os.path.exists(key_path):
        with open(key_path, 'r') as data:
            key = data.read()
    else:
        key = Fernet.generate_key()
        with open(key_path, 'w') as output:
            output.write(key)

    return key

def get_env_key(obj, key=None):
    """
    Return environment variable key to use for lookups within a
    namespace represented by the package name.

    For example, any varialbes for predix.security.uaa are stored
    as PREDIX_SECURITY_UAA_KEY
    """
    return str.join('_', [obj.__module__.replace('.','_').upper(),
        key.upper()])

def get_env_value(obj, attribute):
    """
    Returns the environment variable value for the attribute of
    the given object.

    For example `get_env_value(predix.security.uaa, 'uri')` will
    return value of environment variable PREDIX_SECURITY_UAA_URI.

    """
    varname = get_env_key(obj, attribute)
    var = os.environ.get(varname)
    if not var:
        raise ValueError("%s must be set in your environment." % varname)

    return var

def set_env_value(obj, attribute, value):
    """
    Set the environment variable value for the attribute of the
    given object.

    For example, `set_env_value(predix.security.uaa, 'uri', 'http://...')`
    will set the environment variable PREDIX_SECURITY_UAA_URI to the given
    uri.
    """
    varname = get_env_key(obj, attribute)
    os.environ[varname] = value
    return varname

class PredixCloudRequiredError(Exception):
    """
    A Predix Cloud Requirement Error will be raised when trying to utilize
    a service that is blocked for access anywhere outside of the Predix Cloud.
    """
    def __init__(self, message=None):
        if not message:
            message = "Service only available running in Predix Cloud Foundry environment."

        logging.warning(message)
        super(PredixCloudRequiredError, self).__init__(message)
